/*
 * jBoss, the OpenSource EJB server
 *
 * Distributable under GPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.minerva.datasource;

import java.sql.Connection;
import java.io.PrintWriter;
import java.util.*;
import javax.sql.*;
import javax.naming.*;
import javax.naming.spi.*;
import org.jboss.minerva.factories.XAConnectionFactory;
import org.jboss.minerva.pools.ObjectPool;
import org.jboss.logging.Logger;

/**
 * DataSource for transactional JDBC pools.  This handles configuration
 * parameters for both the pool and the JDBC connection.  It is important that
 * you set all the configuration parameters before you initialize the DataSource,
 * and you must initialize it before you use it.  All the configuration
 * parameters are not documented here; you are instead referred to ObjectPool
 * and XAConnectionFactory.
 * @see org.jboss.minerva.pools.ObjectPool
 * @see org.jboss.minerva.factories.XAConnectionFactory
 * @version $Revision: 1.4 $
 * @author Aaron Mulder (ammulder@alumni.princeton.edu)
 */
public class XAPoolDataSource implements DataSource, Referenceable, ObjectFactory {
    private static HashMap sources = new HashMap();
    /**
     * Gets all the current JDBC pool data sources.
     */
    public static Collection getDataSources() {
        return new HashSet(sources.values());
    }
    /**
     * Gets a specific JDBC pool data source by pool name.
     */
    public static XAPoolDataSource getDataSource(String poolName) {
        return (XAPoolDataSource)sources.get(poolName);
    }


    private ObjectPool pool;
    private XAConnectionFactory factory;
    private PrintWriter logWriter;
    private int timeout;
    private boolean initialized = false;
    private String jndiName;

    /**
     * Creates a new XA pool data source.  Be sure to configure it and then
     * call initialize before you try to use it.
     */
    public XAPoolDataSource() {
        pool = new ObjectPool();
        try {
            factory = new XAConnectionFactory();
        } catch(NamingException e) {
            Logger.exception(e);
        }
        PoolDriver.instance();
    }

// Unique properties
    /**
     * If you use this to set a JNDI name, this pool will be bound to that name
     * using the default InitialContext.  You can also do this manually if you
     * have additional requirements.
     */
    public void setJNDIName(String name) throws NamingException {
        InitialContext ctx = new InitialContext();
        if(jndiName != null && !jndiName.equals(name))
            ctx.unbind(jndiName);
        if(name != null)
            ctx.bind(name, this);
        jndiName = name;
    }

    /**
     * Gets the JNDI name this pool is bound to.  Only valid if you used
     * setJNDIName to bind it.
     * @see #setJNDIName
     */
    public String getJNDIName() {return jndiName;}

// XA properties
    public void setDataSource(XADataSource ds) {factory.setDataSource(ds);}
    public XADataSource getDataSource() {return factory.getDataSource();}
    public void setTransactionManagerJNDIName(String name) {factory.setTransactionManagerJNDIName(name);}
    public String getTransactionManagerJNDIName() {return factory.getTransactionManagerJNDIName();}
    public void setJDBCUser(String user) {factory.setUser(user);}
    public String getJDBCUser() {return factory.getUser();}
    public void setJDBCPassword(String password) {factory.setPassword(password);}
    public String getJDBCPassword() {return factory.getPassword();}
// Pool properties
    public void setPoolName(String name) {
        pool.setName(name);
        sources.put(pool.getName(), this);
    }
    public String getPoolName() {return pool.getName();}
    public void setMinSize(int size) {pool.setMinSize(size);}
    public int getMinSize() {return pool.getMinSize();}
    public void setMaxSize(int size) {pool.setMaxSize(size);}
    public int getMaxSize() {return pool.getMaxSize();}
    public void setBlocking(boolean blocking) {pool.setBlocking(blocking);}
    public boolean isBlocking() {return pool.isBlocking();}
    public void setIdleTimeoutEnabled(boolean allowShrinking) {pool.setIdleTimeoutEnabled(allowShrinking);}
    public boolean isIdleTimeoutEnabled() {return pool.isIdleTimeoutEnabled();}
    public void setGCEnabled(boolean allowGC) {pool.setGCEnabled(allowGC);}
    public boolean isGCEnabled() {return pool.isGCEnabled();}
    public void setMaxIdleTimeoutPercent(float percent) {pool.setMaxIdleTimeoutPercent(percent);}
    public float getMaxIdleTimeoutPercent() {return pool.getMaxIdleTimeoutPercent();}
    public void setIdleTimeout(long millis) {pool.setIdleTimeout(millis);}
    public long getIdleTimeout() {return pool.getIdleTimeout();}
    public void setGCMinIdleTime(long millis) {pool.setGCMinIdleTime(millis);}
    public long getGCMinIdleTime() {return pool.getGCMinIdleTime();}
    public void setGCInterval(long millis) {pool.setGCInterval(millis);}
    public long getGCInterval() {return pool.getGCInterval();}
    public void setInvalidateOnError(boolean invalidate) {pool.setInvalidateOnError(invalidate);}
    public boolean isInvalidateOnError() {return pool.isInvalidateOnError();}
    public void setTimestampUsed(boolean timestamp) {pool.setTimestampUsed(timestamp);}
    public boolean isTimestampUsed() {return pool.isTimestampUsed();}

// Other methods

    /**
     * Initializes the pool.  You need to have configured all the pool and
     * XA properties first.
     */
    public void initialize() {
        initialized = true;
        pool.setObjectFactory(factory);
        pool.initialize();
    }

    /**
     * Returns a string describing the pool status (number of connections
     * created, used, and maximum).
     */
    public String getPoolStatus() {
        return pool.toString();
    }

    /**
     * Shuts down this data source and the underlying pool.  If you used
     * setJNDI name to bind it in JNDI, it is unbound.
     */
    public void close() {
        try {
            setJNDIName(null);
        } catch(NamingException e) {
            Logger.exception(e);
        }
        sources.remove(pool.getName());
        pool.shutDown();
        pool = null;
        factory = null;
    }

    /**
     * Gets a connection from the pool.
     */
    public Connection getConnection() throws java.sql.SQLException {
        if(!initialized) initialize();
        return ((XAConnection)pool.getObject()).getConnection();
    }

    /**
     * Gets a new connection from the pool.  If a new connection must be
     * created, it will use the specified user name and password.  If there is
     * a connection available in the pool, it will be used, regardless of the
     * user name and password use to created it initially.
     */
    public Connection getConnection(String user, String password) throws java.sql.SQLException {
        if(!initialized) initialize();
        factory.setUser(user);
        factory.setPassword(password);
        return ((XAConnection)pool.getObject()).getConnection();
    }

    /**
     * Gets a log writer used to record pool events.
     */
    public PrintWriter getLogWriter() throws java.sql.SQLException {
        return logWriter;
    }

    /**
     * Sets a log writer used to record pool events.
     */
    public void setLogWriter(PrintWriter writer) throws java.sql.SQLException {
        logWriter = writer;
        pool.setLogWriter(writer);
    }

    /**
     * This property is not used by this implementation.
     */
    public int getLoginTimeout() throws java.sql.SQLException {
        return timeout;
    }

    /**
     * This property is not used by this implementation.
     */
    public void setLoginTimeout(int timeout) throws java.sql.SQLException {
        this.timeout = timeout;
    }

    // Referenceable implementation ----------------------------------
    /**
     * Gets a reference to this data source.
     */
    public Reference getReference() {
        return new Reference(getClass().getName(), new StringRefAddr("XAPool", pool.getName()), getClass().getName(), null);
    }

    // ObjectFactory implementation ----------------------------------
    /**
     * Decodes a reference to a specific pool data source.
     */
    public Object getObjectInstance(Object obj, Name name, Context nameCtx,
                                    Hashtable environment) {
        if(obj instanceof Reference) {
            Reference ref = (Reference)obj;
            if(ref.getClassName().equals(getClass().getName())) {
                RefAddr addr = ref.get("XAPool");
                return sources.get((String)addr.getContent());
            }
        }
        return null;
    }
}